# -*- coding: utf-8 -*-
"""Unit test suite for `bottle_jwt` plugin.
"""
from __future__ import unicode_literals
from __future__ import print_function


import pytest
from bottle_jwt import BaseAuthBackend, JWTAuthError, JWTForbiddenError, JWTUnauthorizedError
from bottle_jwt.compat import b
import time


def test_backend_interface_check_pass(backend):
    """Testing for backend interface check pass.
    """
    assert isinstance(backend, BaseAuthBackend)


def test_backend_interface_check_fail():
    """Testing for backend interface check fail.
    """
    class IncompleteBackend(object):
        def get_user(self, user_id):
            pass

    assert not isinstance(IncompleteBackend(), BaseAuthBackend)


def test_provider_authenticate_pass(jwt_provider, request):
    """Testing `bottle_jwt.auth.JWTProvider.authenticate` method pass.
    """

    req = request({'username': 'pav', 'password': '123'})

    assert jwt_provider.authenticate(req)


def test_provider_authenticate_fail(jwt_provider, request):
    """Testing `bottle_jwt.auth.JWTProvider.authenticate` method fail.
    """

    req = request({'username': 'pav', 'password': '1234'})

    with pytest.raises(JWTAuthError):
        jwt_provider.authenticate(req)


def test_provider_authorize_pass(jwt_provider, request):
    """Testing `bottle_jwt.auth.JWTProvider.authorize` method pass.
    """
    req = request({'username': 'pav', 'password': '123'})

    token, _ = jwt_provider.authenticate(req)

    req.set_header('Authorization', 'JWT {}'.format(token.decode("utf-8")))

    assert jwt_provider.authorize(req)['username'] == 'pav'


def test_provider_authorize_fail(jwt_provider, request):
    """Testing `bottle_jwt.auth.JWTProvider.authorize` method pass.
    """
    req = request({'username': 'pav', 'password': '123'})

    token, _ = jwt_provider.authenticate(req)

    # test for invalid token
    req.set_header('Authorization', 'JWT {}'.format(token[:-1].decode("utf-8")))

    with pytest.raises(JWTUnauthorizedError):
        jwt_provider.authorize(req)

    # test for no token at all
    with pytest.raises(JWTForbiddenError):
        jwt_provider.authorize(request({'username': 'pav', 'password': '123'}))

    # test for expired token
    req.set_header('Authorization', 'JWT {}'.format(token.decode("utf-8")))
    time.sleep(2)

    with pytest.raises(JWTUnauthorizedError):
        jwt_provider.authorize(req)


def test_auth_plugin_login_pass(bottle_app):
    """Test `bottle_jwt.JWTProviderPlugin` login web handler pass.
    """

    data = bottle_app.post_json('/auth', {'username': 'pav', 'password': '123'})

    assert 'token' in data.json


def test_auth_plugin_login_fail(bottle_app):
    """Test `bottle_jwt.JWTProviderPlugin` login web handler fail.
    """

    data = bottle_app.post_json('/auth', {'username': 'pav', 'password': '12'})

    assert data.json == {"AuthError": "Unable to authenticate User"}


def test_auth_plugin_authentication_pass(bottle_app):
    """Test `bottle_jwt.JWTProviderPlugin` authentication pass.
    """

    login = bottle_app.post_json('/auth', {'username': 'pav', 'password': '123'}).json

    resource = bottle_app.get(
        '/',
        extra_environ=dict(HTTP_AUTHORIZATION=b('JWT {}'.format(login['token'])))
    )

    assert resource.json


def test_auth_plugin_authentication_fail(bottle_app):
    """Test `bottle_jwt.JWTProviderPlugin` authentication pass.
    """

    login = bottle_app.post_json('/auth', {'username': 'pav', 'password': '123'}).json

    # test with no auth

    response = bottle_app.get('/', expect_errors=True)

    assert response.status == '403 Forbidden'

    # test for expired token

    time.sleep(4)

    response = bottle_app.get(
        '/',
        expect_errors=True,
        extra_environ=dict(HTTP_AUTHORIZATION=b('JWT {}'.format(login['token'])))
    )

    assert response.status == '401 Unauthorized'



# -*- coding: utf-8 -*-
"""`bottle_jwt.auth` module.

Main auth providers class implementation.
"""

from __future__ import unicode_literals
from __future__ import print_function

import abc
import six
from bottle_jwt.compat import signature

__all__ = ['BaseAuthBackend', ]


@six.add_metaclass(abc.ABCMeta)
class BaseAuthBackend(object):
    """Auth Provider Backend Interface. Defines a standard API for implementation
    in order to work with different backends (SQL, Redis, Filesystem-based, external
    API services, etc.)

    Notes:
        It is not necessary to subclass `BaseAuthBackend` in order to make `bottle-jwt` plugin to
        work, as long as you implement it's API. For example all the following examples are valid.

    Examples:
        >>> class DummyExampleBackend(object):
        ...     credentials = ('admin', 'qwerty')
        ...     user_id = 1
        ...
        ...     def authenticate_user(self, username, password):
        ...         if (username, password) == self.credentials
        ...             return {'user': 'admin', 'id': 1}
        ...         return None
        ...
        ...     def get_user(self, user_id):
        ...         return {'user': 'admin '} if user_id == self.user_id else None
        ...
        >>> class SQLAlchemyExampleBackend(object):
        ...     def __init__(self, some_orm_model):
        ...         self.orm_model = some_orm_model
        ...
        ...     def authenticate(self, username, password):
        ...         return self.orm_model.get(email=username, password=password) or None
        ...
        ...     def get_user(self, user_id):
        ...         return self.orm_model.get(id=user_uid) or None
        ...
        """

    @abc.abstractmethod
    def authenticate_user(self, username, password):  # pragma: no cover
        """User authentication method. All subclasses must implement the
        `authenticate_user` method with the following specs.

        Args:
            username (str): User identity for the backend (email/username).
            password (str): User secret password.

        Returns:
            A dict representing User record if authentication is succesful else None.

        Raises:
            `bottle_jwt.error.JWTBackendError` if any exception occurs.
        """
        pass

    @abc.abstractmethod
    def get_user(self, user_id):  # pragma: no cover
        """User data retrieval method. All subclasses must implement the
        `get_user` method with the following specs.

        Args:
            user_id (object): User identity in backend.

        Returns:
            User data (dict) if user exists or None.

        Raises:
            `bottle_jwt.error.JWTBackendError` if any exception occurs.
        """
        pass

    @classmethod
    def __subclasshook__(cls, subclass):
        """Useful for checking interface for backends that don't inherit from
        BaseAuthBackend.
        """
        if cls is BaseAuthBackend:
            try:
                authenticate_user_signature = set(signature(subclass.authenticate_user).parameters)
                get_user_signature = set(signature(subclass.get_user).parameters)

                return authenticate_user_signature.issuperset({'username', 'password'}) and \
                    get_user_signature.issuperset({'user_id'})

            except AttributeError:
                return False

        return NotImplemented  # pragma: no cover



# -*- coding: utf-8 -*-
"""`bottle_jwt.auth` module.

Main auth providers class implementation.
"""

from __future__ import unicode_literals
from __future__ import print_function

import base64
import bottle
import collections
import jwt
import datetime
import logging
from bottle_jwt.compat import signature
from bottle_jwt.backends import BaseAuthBackend
from bottle_jwt.error import JWTBackendError, JWTAuthError, JWTForbiddenError, JWTUnauthorizedError
from bottle_jwt.compat import b


try:
    import ujson as json

except ImportError:
    try:
        import simplejson as json

    except ImportError:
        import json


logger = logging.getLogger(__name__)


auth_fields = collections.namedtuple('AuthField', 'user_id, password')


def jwt_auth_required(callable_obj):
    """A decorator that signs a callable object with an 'auth_required'
    attribute (True). We use this attribute to find which handler callbacks
    require an authorized for protected access.

    Args:
        callable_obj (instance): A handler callable object.

    Returns:
        The callable object.
    """
    setattr(callable_obj, 'auth_required', True)

    return callable_obj


class JWTProvider(object):
    """JWT Auth provider concrete class.
    """

    def __init__(self, fields, backend, secret, id_field='id', algorithm='HS256', ttl=None):
        if not isinstance(backend, BaseAuthBackend):  # pragma: no cover
            raise TypeError('backend instance does not implement {} interface'.format(BaseAuthBackend))

        self.id_field = id_field
        self.user_field = auth_fields(*fields)
        self.secret = secret
        self.backend = backend
        self.algorithm = algorithm
        self.ttl = ttl

    @property
    def expires(self):
        """Computes the token expiration time based on `self.ttl` attribute.
        """
        return datetime.datetime.utcnow() + datetime.timedelta(
            seconds=self.ttl
        )

    def create_token(self, user, ttl=None):
        """Creates a new signed JWT-valid token.

        Args:
            user (dict): The user record in key/value mapping from instance backend.
            ttl (int): Optional time to live value.

        Returns:
            A valid JWT with expiration signature
        """
        user_id = json.dumps(user.get(self.id_field)).encode('utf-8')

        payload = {'sub': base64.b64encode(bytes(user_id)).decode("utf-8")}

        if self.ttl:
            # you can override instance default ttl in special cases.
            if ttl:
                payload['exp'] = datetime.datetime.utcnow() + datetime.timedelta(seconds=ttl)

            else:
                payload['exp'] = self.expires

        logger.debug("Token created for payload: {}".format(str(payload)))

        return jwt.encode(payload, self.secret, algorithm=self.algorithm), payload['exp']

    def validate_token(self, token=''):
        """Validate JWT token.

        Args:
            token (str): A Json Web token string.

        Returns:
            The decrypted token data (dict)

        Raises:
            JWTProviderError, if no token is provided, or if it is expired.
        """
        if not token:
            logger.debug("Forbidden access")
            raise JWTForbiddenError('Cannot access this resource!')

        try:
            decoded = jwt.decode(
                token.split(" ", 1).pop(),
                self.secret,
                algorithms=self.algorithm
            )

            logger.debug("Token validation passed: {}".format(token))

            user_uid = decoded.get('sub')

            if not user_uid:  # pragma: no cover
                raise JWTUnauthorizedError('Invalid User token')

            user = self.backend.get_user(json.loads(base64.b64decode(user_uid).decode('utf-8')))

            if user:
                return user

            raise JWTUnauthorizedError('Invalid User token')

        except (jwt.DecodeError, jwt.ExpiredSignatureError) as e:
            logger.debug("{}: {}".format(e.args[0], token))
            raise JWTUnauthorizedError("Invalid auth token provided.")

    def authenticate(self, request):
        """Returns a valid JWT for provided credentials.

        Args:
            request (instance): bottle.request instance.

        Returns:
            A JWT token string.

        Raises:
            BackendError, if an auth backend error occurs.
            JWTProviderError,, if user can't be authorized.
        """
        if request.content_type.startswith('application/json'):  # pragma: no cover
            try:
                username = request.json.get(self.user_field.user_id)
                password = request.json.get(self.user_field.password)
            except (AttributeError, json.decoder.JSONDecodeError):
                raise JWTAuthError("Unable to authenticate User")

        else:
            username = request.forms.get(self.user_field.user_id)
            password = request.forms.get(self.user_field.password)

        user = self.backend.authenticate_user(username, password)

        if user:
            return self.create_token(user)

        raise JWTAuthError("Unable to authenticate User")

    def authorize(self, request):
        """Checks if incoming request is authenticated.

        Args:
            request (instance): bottle.request instance.

        Returns:
            Request JWT decrypted payload, if request is authenticated else
            False.

        Raises:
            JWTProvider, if no auth header if present or invalid/expired
            token is provided.
        """
        user_token = request.get_header("Authorization", '')
        return self.validate_token(user_token) or False


class JWTProviderPlugin(object):
    """A `bottle.Bottle` application plugin for JWTProvider.

    Attributes:
        keyword (str): The string keyword for application registry.
        provider (instance): A JWTProvider instance.
        login_enable (bool): If True app is mounted with a login handler.
        auth_endpoint (str): The authentication uri for provider if
                             login_enabled is True.
        kwargs : JWTProvider init parameters.
    """
    scope = ('plugin', 'middleware')
    api = 2

    def __init__(self, keyword, auth_endpoint, login_enable=True, scope='plugin', **kwargs):
        self.keyword = keyword
        self.login_enable = login_enable
        self.scope = scope
        self.provider = JWTProvider(**kwargs)
        self.auth_endpoint = auth_endpoint

    def setup(self, app):  # pragma: no cover
        """Make sure that other installed plugins don't affect the same
        keyword argument and check if metadata is available.
        """

        if self.login_enable:

            #  Route a login handler in bottle.py app instance.
            @app.post(self.auth_endpoint)
            def auth_handler():
                try:
                    token, expires = self.provider.authenticate(bottle.request)
                    return {"token": token.decode("utf-8"), "expires": str(expires)}

                except JWTAuthError as error:
                    return {"AuthError": error.args[0]}

                except JWTBackendError:
                    return {"AuthBackendError": "Try later or contact admin!"}

        for other in app.plugins:
            if not isinstance(other, JWTProviderPlugin):
                continue

            if other.keyword == self.keyword:
                raise bottle.PluginError("Found another JWT plugin "
                                         "with conflicting settings ("
                                         "non-unique keyword).")

    def apply(self, callback, context):  # pragma: no cover
        """Implement bottle.py API version 2 `apply` method.
        """

        _signature = signature(callback).parameters

        def injected(*args, **kwargs):
            if self.keyword in _signature:
                kwargs[self.keyword] = self.provider
            return callback(*args, **kwargs)

        def wrapper(*args, **kwargs):
            try:
                user = self.provider.authorize(bottle.request)
                setattr(bottle.request, 'get_user', lambda _: user)
                return injected(*args, **kwargs)

            except JWTUnauthorizedError as error:
                bottle.response.content_type = b('application/json')
                bottle.response._status_line = b('401 Unauthorized')
                return {"AuthError": error.args}

            except JWTForbiddenError as error:
                bottle.response.content_type = b('application/json')
                bottle.response._status_line = b('403 Forbidden')
                return {"AuthError": error.args}

            except JWTBackendError:
                bottle.response.content_type = b('application/json')
                bottle.response._status_line = b('503 Service Unavailable')
                return {"AuthBackendException": "Try later or contact admin!"}

        if self.scope == 'middleware':
            logger.debug("JWT Authentication: {}".format(context.rule))
            return wrapper

        if not hasattr(callback, 'auth_required'):
            return injected

        logger.debug("JWT Authentication: {}".format(context.rule))
        return wrapper



# -*- coding: utf-8 -*-
#
#    Copyright (C) 2015  Papavassiliou Vassilis
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""`bottle-jwt` package.

JSON Web Token Authentication plugin for bottle.py apps.
"""

__author__ = 'Papavassiliou Vassilis'
__date__ = '2017-1-5'
__version__ = '0.11'
__all__ = ['JWTProviderPlugin', 'JWTProvider', 'jwt_auth_required',
           'BaseAuthBackend', 'JWTError', 'JWTBackendError', 'JWTAuthError',
           'JWTForbiddenError', 'JWTUnauthorizedError']

from bottle_jwt.auth import (JWTProvider, JWTProviderPlugin, jwt_auth_required)
from bottle_jwt.backends import BaseAuthBackend
from bottle_jwt.error import (JWTError, JWTBackendError, JWTAuthError,
                              JWTForbiddenError, JWTUnauthorizedError)

import logging


try:  # Python 2.7+
    from logging import NullHandler

except ImportError:  # pragma: no cover
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

logging.getLogger(__name__).addHandler(NullHandler())



# -*- coding: utf-8 -*-
"""`bottle_jwt.compat` module.

Provides python 2, 3 compatibility functions.
"""

from __future__ import unicode_literals
from __future__ import print_function

import sys
import codecs


if sys.version_info.major > 2:
    from inspect import signature as _signature
    signature = _signature

    def b(string):
        return string

else:
    from inspect import getargspec

    class _Signature(object):
        def __init__(self, callable_obj):
            self.spec = getargspec(callable_obj).args

        @property
        def parameters(self):
            return [arg for arg in self.spec if arg != 'self']

    signature = _Signature

    def b(string):
        return codecs.latin_1_encode(string)[0]



from distutils.core import setup
import re
import ast


_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('bottle_jwt/__init__.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')
    ).group(1)))


setup(
    name='bottle-jwt',
    version=version,
    packages=['bottle_jwt'],
    url='',
    license='GLPv3',
    author='Papavassiliou Vassilis',
    author_email='vpapavasil@gmail.com',
    description='JWT Auth plugin for bottle.py applications',
    install_requires=['pyjwt']
)



# -*- coding: utf-8 -*-
"""`bottle_jwt.error` module.

Provides package Exception hierarchy.
"""


class JWTError(Exception):
    """Base package Error class.
    """
    pass


class JWTBackendError(JWTError):
    """Raises when an authentication backend Error occurs.
    """
    pass


class JWTAuthError(JWTError):
    """Raises when an authentication provider error occurs.
    """
    pass


class JWTUnauthorizedError(JWTAuthError):
    """Raises when Unauthorized access occurs.
    """
    pass


class JWTForbiddenError(JWTAuthError):
    """Raises when no authorized access requests protected resource.
    """
    pass
